"""Utility functions and classes used internally by pybi."""

import json
from pathlib import Path
import re

from requests import Response, Session
from requests.exceptions import HTTPError


def to_snake_case(s):
    pattern = re.compile("((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))")
    return pattern.sub(r"_\1", s).lower()


def to_camel_case(text):
    s = text.replace("-", " ").replace("_", " ")
    s = s.split()

    if len(text) == 0:
        return text

    return s[0] + "".join(i.capitalize() for i in s[1:])


def remove_no_values(
    d: dict,
) -> dict:
    """
    Recursively remove keys with a value of `None` or `{}` from a dictionary.
    If the result of removing `None` is `{}`, then this is removed until
    only keys with values remain.

    Parameters
    ----------
    `d` : `dict`
        dict to remove None or empty dicts from.

    Returns
    -------
    `dict`
        dict with `None` or `{}` removed.
    """

    new_d = {}

    for k, v in d.items():
        if isinstance(v, dict):
            v = remove_no_values(v)
        if not v in (None, {}):
            new_d[k] = v

    return new_d


def file_path_from_components(
    file_name: str,
    extension: str,
    directory: str | Path = None,
):
    """
    Create a file path from file path components.

    Parameters
    ----------
    `file_name` : `str`
        Name of the file (without extension).
    `extension` : `str`
        The file extension.
    `directory` : `str | Path`, optional
        Directory of the file.

    Returns
    -------
    `Path`
        File path to the file.

    """

    if isinstance(directory, str):
        f_dir = Path(directory)
    elif directory is None:
        f_dir = Path()  # cwd
    else:
        f_dir = directory
    
    # Be a bit more flexible and accept .ext
    if extension[0] == ".":
        f_ext = extension[1:]
    else:
        f_ext = extension

    return f_dir / f"{file_name}.{f_ext}"


class RequestsMixin:
    """
    Mixin class for http requests and response handling.

    Lets the PowerBI client and Resources use the same functionality for
    making requests and handling the response.

    Mostly requests lib boilerplate with a sprinkling of json parsing.
    """

    def delete(
        self,
        resource: str,
        session: Session,
        success_codes: list[int] = [200, 201],
    ) -> None:
        """
        Make a delete request to an api endpoint.

        Parameters
        ----------
        `resource` : `str`
            URL of the resource.
        `session` : `Session`
            Requests Session object used to make the request.
        `success_codes` : `list[int]`, optional
            HTTP response status codes that indicate a successful request.
            Status codes not equal to these will raise an `HTTPError`.

        Raises
        ------
        `HTTPError`
            If the response status code was not found in `success_codes`.
        """

        response = session.delete(resource)

        if response.status_code not in success_codes:
            raise HTTPError(
                f"""Encountered api error. Response: 
                
                {json.dumps(response.json(), indent=True)})"""
            )

    def patch(
        self,
        resource: str,
        session: Session,
        payload: dict = None,
        success_codes: list[int] = [200, 201],
    ) -> Response:
        """
        Patch a resource.

        Parameters
        ----------
        `resource` : `str`
            URL of the resource to patch.
        `session` : `Session`
            Requests Session object used to make the request.
        `payload` : `dict`
            Data to patch on the resource.
        `success_codes` : `list[int]`, optional
            HTTP response status codes that indicate a successful request.
            Status codes not equal to these will raise an `HTTPError`.

        Returns
        -------
        `Response`
            Response generated from the patch request.

        Raises
        ------
        `HTTPError`
            If the response status code was not found in `success_codes`.
        """

        response = session.patch(resource, json=payload)

        if response.status_code not in success_codes:
            raise HTTPError(
                f"Encountered api error. Response: {json.dumps(response.json(), indent=True)})"
            )

        return response

    def post(
        self,
        resource: str,
        session: Session,
        payload: dict = None,
        success_codes: list[int] = [200, 201, 202],
    ) -> Response:
        """
        Post data to an api endpoint.

        Parameters
        ----------
        `resource` : `str`
            URL of the resource to post to.
        `session` : `Session`
            Requests Session object used to make the request.
        `payload` : `dict`
            Data to post to the resource.
        `success_codes` : `list[int]`, optional
            HTTP response status codes that indicate a successful request.
            Status codes not equal to these will raise an `HTTPError`.

        Returns
        -------
        `Response`
            Response generated from the post request.

        Raises
        ------
        `HTTPError`
            If the response status code was not found in `success_codes`.
        """

        response = session.post(resource, json=payload)

        if response.status_code not in success_codes:
            raise HTTPError(
                f"Encountered api error. Response: {json.dumps(response.json(), indent=True)})"
            )

        return response

    def post_raw(
        self,
        resource: str,
        session: Session,
        payload: dict,
        **kwargs: dict,
    ) -> dict:
        """
        Make a post request and return any resulting json.

        Parameters
        ----------
        `resource` : `str`
            URL of the resource to post to.
        `session` : `Session`
            Requests Session to use to make the request.
        `payload` : `dict`
            Request data.

        Returns
        -------
        `dict`
            Request json. `response.json()`

        Raises
        ------
        `Exception`
            If encountered error with the request.
        """

        try:
            response = self.post(resource, session, payload, **kwargs)
            raw = response.json()

            return self.parse_raw(raw)

        except Exception as e:
            raise e

    def put(
        self,
        resource: str,
        session: Session,
        payload: dict,
        success_codes: list[int] = [200, 201],
    ) -> Response:
        """
        Make a put request to an api endpoint.

        Parameters
        ----------
        `resource` : `str`
            URL of the resource.
        `session` : `Session`
            Requests Session object used to make the request.
        `payload` : `dict`
            Payload data to include with the put request.
        `success_codes` : `list[int]`, optional
            HTTP response status codes that indicate a successful request.
            Status codes not equal to these will raise an `HTTPError`.

        Returns
        -------
        `Response`
            requests Response object.

        Raises
        ------
        `HTTPError`
            If the response status code was not found in `success_codes`.
        """

        response = session.put(resource, json=payload)

        if response.status_code not in success_codes:
            raise HTTPError(
                f"""Encountered api error. Response: 
                
                {json.dumps(response.json(), indent=True)})"""
            )

        return response

    def get(
        self,
        resource: str,
        session: Session,
        params: dict = None,
        success_codes: list[int] = [200, 201],
    ) -> Response:
        """
        Get a resource from an api endpoint.

        Parameters
        ----------
        `resource` : `str`
            URL of the resource.
        `session` : `Session`
            Requests Session object used to make the request.
        `params` : `dict`, optional
            Request parameters.
        `success_codes` : `list[int]`, optional
            HTTP response status codes that indicate a successful request.
            Status codes not equal to these will raise an `HTTPError`.

        Returns
        -------
        `Response`
            requests Response object.

        Raises
        ------
        `HTTPError`
            If the response status code was not found in `success_codes`.
        """

        response = session.get(resource, params=params)

        if response.status_code not in success_codes:
            raise HTTPError(
                f"""Encountered api error. Response: 
                
                {json.dumps(response.json(), indent=True)})"""
            )

        return response

    def get_raw(
        self,
        resource: str,
        session: Session,
        params: dict = None,
        **kwargs: dict,
    ) -> dict | list:
        """
        Request an api resource, parse the response json, and return the
        parsed json.

        Parameters
        ----------
        `resource` : `str`
            URL of the api resource.
        `session` : `Session`
            Requests Session object used to make the request.

        Returns
        -------
        `dict | list`
            Returns either the dict representation of the resource, or list
            of resources.

        Raises
        ------
        `Exception`
            Error encountered during the request process.
        """

        try:
            response = self.get(resource, session, params, **kwargs)
            raw = response.json()

            return self.parse_raw(raw)

        except Exception as e:
            raise e

    def parse_raw(
        self,
        raw: dict,
    ) -> dict | list[dict]:
        """
        Parses json from an api response.

        Depending on the endpoint, the json will either include a dict
        representation of a single object, or a list of dicts found in
        the "values" key. If a list, this method will return the list,
        rather than the original json dict.

        Parameters
        ----------
        `raw` : `dict`
            Raw json from the api response.

        Returns
        -------
        `dict | list[dict]`
            Either dict representation of the object, or a list of dict
            representations.
        """

        if "value" in raw:
            return raw["value"]
        else:
            return raw



"""
Module implements a wrapper around the Power BI Rest API.

Users construct an instance of the PowerBI client and call its methods.

Full API documentation can be found at:

https://learn.microsoft.com/en-us/rest/api/power-bi/
"""

import requests
from requests.exceptions import HTTPError

from pbipy import settings
from pbipy.groups import Group
from pbipy.datasets import Dataset
from pbipy.reports import Report
from pbipy.utils import RequestsMixin, remove_no_values


class PowerBI(RequestsMixin):
    """
    User Interface into Power BI Rest Api.

    Users interact with their Power BI service by constructing an instance
    of this object and calling its methods.

    Authentication with the Power BI service requires a `bearer_token` which
    must be generated in advance before creating the `PowerBI` object.
    How you the token is generated depends on the user's Azure and Power
    BI configuration.

    In general, `PowerBI()` methods (mostly) follow the operations described
    here:

    https://learn.microsoft.com/en-us/rest/api/power-bi/

    Parameters
    ----------
    `bearer_token` : `str`
        Bearer token used to authenticate with your Power BI service.

    """

    BASE_URL = settings.BASE_URL

    def __init__(
        self,
        bearer_token: str,
        session: requests.Session = None,
    ) -> None:
        self.bearer_token = bearer_token

        if not session:
            self.session = requests.Session()

        self.session.headers.update({"Authorization": f"Bearer {self.bearer_token}"})

    # TODO: Add support for passing in a group obj
    def dataset(
        self,
        dataset: str | Dataset,
        group: str = None,
    ) -> Dataset:
        """
        Return the specified dataset from MyWorkspace or the specified
        group.

        Parameters
        ----------
        `dataset` : `str | Dataset`
            Dataset Id of the dataset to retrieve.
        `group` : `str`, optional
            Group Id where the dataset resides., by default None

        Returns
        -------
        `Dataset`
            The specified dataset.

        """

        if isinstance(dataset, Dataset):
            return dataset

        dataset = Dataset(dataset, self.session, group_id=group)
        dataset.load()

        return dataset

    def datasets(
        self,
        group: str = None,
    ) -> list[Dataset]:
        """
        Returns a list of datasets from MyWorkspace or the specified group.

        Parameters
        ----------
        `group` : str, optional
            Group Id where the datasets reside. If not supplied, then datasets
            will be retrieved from MyWorkspace.

        Returns
        -------
        `list[Dataset]`
            List of datasets for MyWorkspace or the specified group.

        """

        if group:
            path = f"/groups/{group}/datasets"
        else:
            path = "/datasets"

        resource = self.BASE_URL + path
        raw = self.get_raw(resource, self.session)

        datasets = [
            Dataset(
                dataset_js.get("id"),
                self.session,
                group_id=group,
                raw=dataset_js,
            )
            for dataset_js in raw
        ]

        return datasets

    def delete_dataset(
        self,
        dataset: str | Dataset,
        group: str = None,
    ) -> None:
        """
        Delete the specified dataset from MyWorkspace or the specified
        group.

        Parameters
        ----------
        `dataset` : `str | Dataset`
            Dataset Id or `Dataset` object to delete.
        `group` : `str`, optional
            Group Id where the dataset resides., by default None

        Raises
        ------
        `HTTPError`
            If the api response status code is not equal to 200.

        """
        # FIXME: Doesn't use the group from the Dataset.

        if isinstance(dataset, Dataset):
            dataset_id = dataset.id
        else:
            dataset_id = dataset

        if group:
            path = f"/groups/{group}/datasets/{dataset_id}"
        else:
            path = f"/datasets/{dataset_id}"

        resource = self.BASE_URL + path

        self.delete(resource, self.session)

    def group(
        self,
        group_id: str,
    ) -> Group:
        """
        Return the specified group.

        Convenience function that is the equivalent of
        `group(filter="id eq 'group_id'")`.

        Parameters
        ----------
        `group_id` : `str`
            Group Id of the group to retrieve.

        Returns
        -------
        `Group`
            The specified group.

        Raises
        ------
        `ValueError`
            If the Group was not found by the api.

        """

        id_filter = f"id eq '{group_id}'"

        groups = self.groups(filter=id_filter)

        if groups == []:
            raise ValueError(f"Group Id: {id_filter}, was not found by the API.")

        return groups[0]

    def groups(
        self,
        filter: str = None,
        skip: int = None,
        top: int = None,
    ) -> list[Group]:
        """
        Returns a list of workspaces the user has access to.

        Parameters
        ----------
        `filter` : `str`, optional
            Filters the results, based on a boolean condition, e.g.,
            `contains(name, 'marketing')`, `name eq 'contoso'`
        `skip` : `int`, optional
            Skips the first n results.
        `top` : `int`, optional
            Returns only the first n results.

        Returns
        -------
        `list[Group]`
            List of `Group` objects the user has access to, and/or
            that matched the specified filters.
        
        Notes
        -----
        See below for more details on filter syntax:
        https://learn.microsoft.com/en-us/power-bi/collaborate-share/service-url-filters

        """

        params = {
            "$filter": filter,
            "$skip": skip,
            "$top": top,
        }

        preparred_params = remove_no_values(params)

        path = self.BASE_URL + "/groups"
        raw = self.get_raw(path, self.session, params=preparred_params)

        groups = [
            Group(
                group_js.get("id"),
                self.session,
                raw=group_js,
            )
            for group_js in raw
        ]

        return groups

    def create_group(
        self,
        name: str,
        workspace_v2: bool = None,
    ) -> Group:
        """
        Create a new workspace.

        Parameters
        ----------
        `name` : `str`
            The name for the new workspace.
        `workspace_v2` : `bool`, optional
            (Preview feature) Whether to create a workspace. The only supported value is `true`.

        Returns
        -------
        `Group`
            The newly created workspace (group).

        """

        payload = {"name": name}

        if workspace_v2 is not None:
            resource = self.BASE_URL + f"/groups?workspaceV2={workspace_v2}"
        else:
            resource = self.BASE_URL + "/groups"

        raw = self.post_raw(resource, self.session, payload)
        # Endpoint returns as list with one element.
        group_raw = raw[0]

        return Group(group_raw.get("id"), self.session, raw=group_raw)

    def delete_group(
        self,
        group: str | Group,
    ) -> None:
        """
        Delete the specified workgroup.

        Parameters
        ----------
        `group` : `str | Group`
            Group Id or `Group` object to delete.

        """

        if isinstance(group, Group):
            group_id = group.id
        else:
            group_id = group

        resource = self.BASE_URL + f"/groups/{group_id}"

        self.delete(resource, self.session)

    def report(
        self,
        report: str | Report,
        group: str | Group = None,
    ) -> Report:
        """
        Return the specified report from MyWorkspace or the specified group.

        Parameters
        ----------
        report : `str | Report`
            Report Id to retrieve.
        group : `str | Group`, optional
            Group Id or `Group` object where the report resides.

        Returns
        -------
        `Report`
            The specified `Report` object.

        """

        if isinstance(report, Report):
            return report

        if isinstance(group, Group):
            group_id = group.id
        else:
            group_id = group

        report = Report(report, self.session, group_id=group_id)
        report.load()

        return report

    def reports(
        self,
        group: str | Group = None,
    ) -> list[dict]:
        """
        Return a list of reports for MyWorkspace or the specified
        group (workspace).

        Parameters
        ----------
        `group` : `str | Group`, optional
            Group Id or `Group` object where the reports reside., by default None

        Returns
        -------
        `list[dict]`
            List of `Report` objects for MyWorkspace or the specified
            group (workspace).

        """

        if isinstance(group, Group):
            group_id = group.id
        else:
            group_id = group

        if group_id:
            path = f"/groups/{group_id}/reports"
        else:
            path = "/reports"

        resource = self.BASE_URL + path
        raw = self.get_raw(resource, self.session)

        reports = [
            Report(
                report_js.get("id"),
                self.session,
                group_id=group_id,
                raw=report_js,
            )
            for report_js in raw
        ]

        return reports

    def delete_report(
        self,
        report: str | Report,
        group: str | Group = None,
    ) -> None:
        """
        Delete the specified report. 
        
        If a `Report` type is provided as `report`, then it's `group_id` 
        will be taken as the `group` value.

        Parameters
        ----------
        `report` : `str | Report`
            Report Id or `Report` object to delete.
        `group` : `str | Group`, optional
            The Group Id or `Group` object where the report resides. If
            a `Report` object is provided then the `group_id` of that report
            will override the provided group value.

            If no group value is provided, it is assumed the report resides
            in the current user's workspace.
        
        """

        if isinstance(report, Report):
            report_id = report.id
        else:
            report_id = report
        
        # If we got report, use its group_id and ignore provided
        if isinstance(report, Report):
            group_id = report.group_id
        else:
            if isinstance(group, Group):
                group_id = group.id
            else:
                group_id = group
        
        if group_id:
            resource = self.BASE_URL + f"/groups/{group_id}/reports/{report_id}"
        else:
            resource = self.BASE_URL + f"/reports/{report_id}"
        
        self.delete(resource, self.session)



from pbipy.resources import Resource
from pbipy.utils import remove_no_values


from requests import Session


class Group(Resource):
    _REPR = [
        "id",
        "name",
    ]

    def __init__(
        self,
        id: str,
        session: Session,
        raw: dict = None,
        **kwargs,
    ) -> None:
        super().__init__(id, session, **kwargs)

        self.base_path = f"{self.BASE_URL}/groups/{self.id}"

        if raw:
            self._load_from_raw(raw)

    def add_user(
        self,
        identifier: str,
        principal_type: str,
        access_right: str,
        display_name: str = None,
        email_address: str = None,
        graph_id: str = None,
        profile: dict = None,
        user_type: str = None,
    ) -> None:
        """
        Grants the specified user the specified permissions to the workspace.
        
        Parameters
        ----------
        `identifier` : `str`
            Identifier of the principal.
        `principal_type` : `str`
            The principal type, e.g., "App", "Group", "None", or "User".
        `access_right` : `str`
            The access right (permission level) that a user has on the
            workspace, e.g., "Admin", "Contributor", "Member", "None",
            or "Viewer".
        `display_name` : `str`, optional
            Display name of the principal.
        `email_address` : `str`, optional
            Email address of the user.
        `graph_id` : `str`, optional
            Identifier of the principal in Microsoft Graph. Only available
            for admin APIs.
        `profile` : `dict`, optional
            A Power BI service principal profile. Only relevant for Power
            BI Embedded multi-tenancy solution.
        `user_type` : `str`, optional
            Type of the user.
        
        """

        payload = {
            "identifier": identifier,
            "groupUserAccessRight": access_right,
            "principalType": principal_type,
            "displayName": display_name,
            "emailAddress": email_address,
            "graphId": graph_id,
            "profile": profile,
            "userType": user_type,
        }

        prepared_payload = remove_no_values(payload)
        resource = self.base_path + "/users"

        self.post(resource, self.session, prepared_payload)

    def delete_user(
        self,
        user: str,
        profile: str = None,
    ) -> None:
        """
        Deletes the specified user permissions from the specified workspace.
        
        Parameters
        ----------
        `user` : `str`
            The email address of the user or object ID of the service principal
            to delete.
        `profile` : `str`, optional
            The service principal profile ID to delete.
        
        """

        if profile:
            resource = self.base_path + f"/users/{user}?profileId={profile}"
        else:
            resource = self.base_path + f"/users/{user}"

        self.delete(resource, self.session)

    def update_user(
        self,
        identifier: str,
        principal_type: str,
        access_right: str,
        display_name: str = None,
        email_address: str = None,
        graph_id: str = None,
        profile: dict = None,
        user_type: str = None,
    ) -> None:
        """
        Updates the specified user permissions on the workspace.
        
        Parameters
        ----------
        `identifier` : `str`
            Identifier of the principal.
        `principal_type` : `str`
            The principal type, e.g., "App", "Group", "None", or "User".
        `access_right` : `str`
            The access right (permission level) that a user has on the
            workspace, e.g., "Admin", "Contributor", "Member", "None",
            or "Viewer".
        `display_name` : `str`, optional
            Display name of the principal.
        `email_address` : `str`, optional
            Email address of the user.
        `graph_id` : `str`, optional
            Identifier of the principal in Microsoft Graph. Only available
            for admin APIs.
        `profile` : `dict`, optional
            A Power BI service principal profile. Only relevant for Power
            BI Embedded multi-tenancy solution.
        `user_type` : `str`, optional
            Type of the user.
        
        """

        payload = {
            "identifier": identifier,
            "groupUserAccessRight": access_right,
            "principalType": principal_type,
            "displayName": display_name,
            "emailAddress": email_address,
            "graphId": graph_id,
            "profile": profile,
            "userType": user_type,
        }

        prepared_payload = remove_no_values(payload)
        resource = self.base_path + "/users"

        self.put(resource, self.session, prepared_payload)

    def users(
        self,
        skip: int = None,
        top: int = None,
    ) -> list[dict]:
        """
        Returns a list of users that have access to the workspace.
        
        Parameters
        ----------
        `skip` : `int`, optional
            Skips the first n results.
        `top` : `int`, optional
            Returns onl the first n results.
        Returns
        -------
        `list[dict]`
            List of users that have access to the workspace.
        
        """

        params = {
            "$skip": skip,
            "$top": top,
        }

        prepared_params = remove_no_values(params)
        resource = self.base_path + "/users"

        raw = self.get_raw(resource, self.session, params=prepared_params)

        return raw